// -*- C++ -*-
//===----------------------------------------------------------------------===//
//
// Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===----------------------------------------------------------------------===//

export {
  using ::isalnum _LIBCPP_USING_IF_EXISTS;
  using ::isalpha _LIBCPP_USING_IF_EXISTS;
  using ::isblank _LIBCPP_USING_IF_EXISTS;
  using ::iscntrl _LIBCPP_USING_IF_EXISTS;
  using ::isdigit _LIBCPP_USING_IF_EXISTS;
  using ::isgraph _LIBCPP_USING_IF_EXISTS;
  using ::islower _LIBCPP_USING_IF_EXISTS;
  using ::isprint _LIBCPP_USING_IF_EXISTS;
  using ::ispunct _LIBCPP_USING_IF_EXISTS;
  using ::isspace _LIBCPP_USING_IF_EXISTS;
  using ::isupper _LIBCPP_USING_IF_EXISTS;
  using ::isxdigit _LIBCPP_USING_IF_EXISTS;
  using ::tolower _LIBCPP_USING_IF_EXISTS;
  using ::toupper _LIBCPP_USING_IF_EXISTS;
} // export
