// -*- C++ -*-
//===----------------------------------------------------------------------===//
//
// Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===----------------------------------------------------------------------===//

export namespace std {
  using std::isalnum _LIBCPP_USING_IF_EXISTS;
  using std::isalpha _LIBCPP_USING_IF_EXISTS;
  using std::isblank _LIBCPP_USING_IF_EXISTS;
  using std::iscntrl _LIBCPP_USING_IF_EXISTS;
  using std::isdigit _LIBCPP_USING_IF_EXISTS;
  using std::isgraph _LIBCPP_USING_IF_EXISTS;
  using std::islower _LIBCPP_USING_IF_EXISTS;
  using std::isprint _LIBCPP_USING_IF_EXISTS;
  using std::ispunct _LIBCPP_USING_IF_EXISTS;
  using std::isspace _LIBCPP_USING_IF_EXISTS;
  using std::isupper _LIBCPP_USING_IF_EXISTS;
  using std::isxdigit _LIBCPP_USING_IF_EXISTS;
  using std::tolower _LIBCPP_USING_IF_EXISTS;
  using std::toupper _LIBCPP_USING_IF_EXISTS;
} // namespace std
